@@ -18,6 +18,8 @@
  */
 
 
+#include <string.h>
+
 #include "allegro.h"
 #include "allegro/internal/aintern.h"
 
@@ -274,6 +276,77 @@ static void read_24bit_line(int length, PACKFILE *f, BITMAP *bmp, int line)
 
 
 
+/* read_32bit_line:
+ *  Support function for reading the 32 bit bitmap file format, doing
+ *  our best to convert it down to a 256 color palette.
+ */
+static void read_32bit_line(int length, PACKFILE *f, BITMAP *bmp, int line)
+{
+   int i;
+   RGB c;
+   char a;
+
+   for (i=0; i<length; i++) {
+      c.b = pack_getc(f);
+      c.g = pack_getc(f);
+      c.r = pack_getc(f);
+      a = pack_getc(f);
+      WRITE3BYTES(bmp->line[line]+i*4, makeacol32(c.r, c.g, c.b, a));
+   }
+}
+
+
+
+/* read_bitfields_image:
+ *  For reading the bitfield compressed BMP image format.
+ */
+static void read_bitfields_image(PACKFILE *f, BITMAP *bmp, AL_CONST BITMAPINFOHEADER *infoheader)
+{
+   int k, i;
+   int bpp;
+   int bytes_per_pixel;
+   int red, grn, blu;
+   unsigned long buffer;
+
+   bpp = bitmap_color_depth(bmp);
+   bytes_per_pixel = BYTES_PER_PIXEL(bpp);
+
+   for (i=0; i<(int)infoheader->biHeight; i++) {
+      for (k=0; k<(int)infoheader->biWidth; k++) {
+
+	 pack_fread(&buffer, bytes_per_pixel, f);
+
+	 if (bpp == 15) {
+	    red = (buffer >> 10) & 0x1f;
+	    grn = (buffer >> 5) & 0x1f;
+	    blu = (buffer) & 0x1f;
+	    buffer = (red << _rgb_r_shift_15) |
+		     (grn << _rgb_g_shift_15) |
+		     (blu << _rgb_b_shift_15);
+	 }
+	 else if (bpp == 16) {
+	    red = (buffer >> 11) & 0x1f;
+	    grn = (buffer >> 5) & 0x3f;
+	    blu = (buffer) & 0x1f;
+	    buffer = (red << _rgb_r_shift_16) |
+		     (grn << _rgb_g_shift_16) |
+		     (blu << _rgb_b_shift_16);
+	 }
+	 else {
+	    red = (buffer >> 16) & 0xff;
+	    grn = (buffer >> 8) & 0xff;
+	    blu = (buffer) & 0xff;
+	    buffer = (red << _rgb_r_shift_32) |
+		     (grn << _rgb_g_shift_32) |
+		     (blu << _rgb_b_shift_32);
+	 }
+
+	 memcpy(&bmp->line[(infoheader->biHeight - i) - 1][k * bytes_per_pixel], &buffer, bytes_per_pixel);
+      }
+   }
+}
+
+
 /* read_image:
  *  For reading the noncompressed BMP image format.
  */
@@ -301,6 +374,10 @@ static void read_image(PACKFILE *f, BITMAP *bmp, AL_CONST BITMAPINFOHEADER *info
 	 case 24:
 	    read_24bit_line(infoheader->biWidth, f, bmp, infoheader->biHeight-i-1);
 	    break;
+
+	 case 32:
+	    read_32bit_line(infoheader->biWidth, f, bmp, infoheader->biHeight-i-1);
+	    break;
       }
    }
 }
@@ -463,17 +540,42 @@ static void read_RLE4_compressed_image(PACKFILE *f, BITMAP *bmp, AL_CONST BITMAP
  *  Thanks to Seymour Shlien for contributing this function.
  */
 BITMAP *load_bmp(AL_CONST char *filename, RGB *pal)
+{
+   PACKFILE *f;
+   BITMAP *bmp;
+   ASSERT(filename);
+
+   f = pack_fopen(filename, F_READ);
+   if (!f)
+      return NULL;
+
+   bmp = load_bmp_pf(f, pal);
+
+   pack_fclose(f);
+
+   return bmp;
+}
+
+
+
+/* load_bmp_pf:
+ *  Like load_bmp, but starts loading from the current place in the PACKFILE
+ *  specified. If successful the offset into the file will be left just after
+ *  the image data. If unsuccessful the offset into the file is unspecified,
+ *  i.e. you must either reset the offset to some known place or close the
+ *  packfile. The packfile is not closed by this function.
+ */
+BITMAP *load_bmp_pf(PACKFILE *f, RGB *pal)
 {
    BITMAPFILEHEADER fileheader;
    BITMAPINFOHEADER infoheader;
-   PACKFILE *f;
    BITMAP *bmp;
    PALETTE tmppal;
    int want_palette = TRUE;
    int ncol;
    unsigned long biSize;
    int bpp, dest_depth;
-   ASSERT(filename);
+   ASSERT(f);
 
    /* we really need a palette */
    if (!pal) {
@@ -481,12 +583,7 @@ BITMAP *load_bmp(AL_CONST char *filename, RGB *pal)
       pal = tmppal;
    }
 
-   f = pack_fopen(filename, F_READ);
-   if (!f)
-      return NULL;
-
    if (read_bmfileheader(f, &fileheader) != 0) {
-      pack_fclose(f);
       return NULL;
    }
 
@@ -494,37 +591,61 @@ BITMAP *load_bmp(AL_CONST char *filename, RGB *pal)
 
    if (biSize == WININFOHEADERSIZE) {
       if (read_win_bminfoheader(f, &infoheader) != 0) {
-	 pack_fclose(f);
 	 return NULL;
       }
       /* compute number of colors recorded */
       ncol = (fileheader.bfOffBits - 54) / 4;
-      read_bmicolors(ncol, pal, f, 1);
+
+      if (infoheader.biCompression != BI_BITFIELDS)
+	 read_bmicolors(ncol, pal, f, 1);
    }
    else if (biSize == OS2INFOHEADERSIZE) {
       if (read_os2_bminfoheader(f, &infoheader) != 0) {
-	 pack_fclose(f);
 	 return NULL;
       }
       /* compute number of colors recorded */
       ncol = (fileheader.bfOffBits - 26) / 3;
-      read_bmicolors(ncol, pal, f, 0);
+
+      if (infoheader.biCompression != BI_BITFIELDS)
+	 read_bmicolors(ncol, pal, f, 0);
    }
    else {
-      pack_fclose(f);
       return NULL;
    }
 
    if (infoheader.biBitCount == 24)
       bpp = 24;
+   else if (infoheader.biBitCount == 16)
+      bpp = 16;
+   else if (infoheader.biBitCount == 32)
+      bpp = 32;
    else
       bpp = 8;
 
+   if (infoheader.biCompression == BI_BITFIELDS) {
+      unsigned long redMask = pack_igetl(f);
+      unsigned long grnMask = pack_igetl(f);
+      unsigned long bluMask = pack_igetl(f);
+
+      (void)grnMask;
+
+      if ((bluMask == 0x001f) && (redMask == 0x7C00))
+	 bpp = 15;
+      else if ((bluMask == 0x001f) && (redMask == 0xF800))
+	 bpp = 16;
+      else if ((bluMask == 0x0000FF) && (redMask == 0xFF0000))
+	 bpp = 32;
+      else {
+	 /* Unrecognised bit masks/depth, refuse to load. */
+	 return NULL;
+      }
+   }
+
+
    dest_depth = _color_load_depth(bpp, FALSE);
 
    bmp = create_bitmap_ex(bpp, infoheader.biWidth, infoheader.biHeight);
    if (!bmp) {
-      pack_fclose(f);
       return NULL;
    }
 
@@ -544,6 +665,10 @@ BITMAP *load_bmp(AL_CONST char *filename, RGB *pal)
 	 read_RLE4_compressed_image(f, bmp, &infoheader);
 	 break;
 
+      case BI_BITFIELDS:
+	 read_bitfields_image(f, bmp, &infoheader);
+	 break;
+
       default:
 	 destroy_bitmap(bmp);
 	 bmp = NULL;
@@ -554,14 +679,15 @@ BITMAP *load_bmp(AL_CONST char *filename, RGB *pal)
       if ((bpp != 8) && (!want_palette))
 	 pal = NULL;
 
-      bmp = _fixup_loaded_bitmap(bmp, pal, dest_depth);
+      if (bmp) {
+	 bmp = _fixup_loaded_bitmap(bmp, pal, dest_depth);
+      }
    }
    
    /* construct a fake palette if 8-bit mode is not involved */
    if ((bpp != 8) && (dest_depth != 8) && want_palette)
       generate_332_palette(pal);
 
-   pack_fclose(f);
    return bmp;
 }
 
@@ -574,6 +700,30 @@ BITMAP *load_bmp(AL_CONST char *filename, RGB *pal)
 int save_bmp(AL_CONST char *filename, BITMAP *bmp, AL_CONST RGB *pal) 
 {
    PACKFILE *f;
+   int ret;
+   ASSERT(filename);
+
+   f = pack_fopen(filename, F_WRITE);
+   if (!f) 
+      return -1;
+
+   ret = save_bmp_pf(f, bmp, pal);
+
+   pack_fclose(f);
+
+   return ret;
+}
+
+
+
+/* save_bmp_pf:
+ *  Like save_bmp but writes into the PACKFILE given instead of a new file.
+ *  The packfile is not closed after writing is completed. On success the
+ *  offset into the file is left after the TGA file just written. On failure
+ *  the offset is left at the end of whatever incomplete data was written.
+ */
+int save_bmp_pf(PACKFILE *f, BITMAP *bmp, AL_CONST RGB *pal) 
+{
    PALETTE tmppal;
    int bfSize;
    int biSizeImage;
@@ -581,7 +731,7 @@ int save_bmp(AL_CONST char *filename, BITMAP *bmp, AL_CONST RGB *pal)
    int bpp;
    int filler;
    int c, i, j;
-   ASSERT(filename);
+   ASSERT(f);
    ASSERT(bmp);
 
    depth = bitmap_color_depth(bmp);
@@ -604,10 +754,6 @@ int save_bmp(AL_CONST char *filename, BITMAP *bmp, AL_CONST RGB *pal)
       bfSize = 54 + biSizeImage;       /* header + image data */
    }
 
-   f = pack_fopen(filename, F_WRITE);
-   if (!f) 
-      return -1;
-
    *allegro_errno = 0;
 
    /* file_header */
@@ -667,8 +813,6 @@ int save_bmp(AL_CONST char *filename, BITMAP *bmp, AL_CONST RGB *pal)
 	 pack_putc(0, f);
    }
 
-   pack_fclose(f);
-
    if (*allegro_errno)
       return -1;
    else
